@@ -50,9 +50,10 @@ Finally, the help documentation for your commands is just the __doc__
 string of the function.
 """
 
+import inspect
 import re
 import sys
-import inspect
+import traceback
 
 
 S_IP_ADDRESS = lambda x, token: ['ip_address', token]
@@ -60,7 +61,7 @@ S_WORD = lambda x, token:  ['word', token]
 S_EMAIL_ADDR = lambda x, token:  ['email', token]
 S_OPTION = lambda x, token:  ['option', token.split("-")[-1]]
 S_INT = lambda x, token:  ['int', int(token) ]
-S_BOOL = lambda x, token:  ['bool', token in ['True', 'true', 'yes']]
+S_BOOL = lambda x, token:  ['bool', token in ['True', 'true', 'yes', 't', 'y']]
 S_EMPTY = lambda x, token:  ['empty', '']
 S_STRING = lambda x, token:  ['string', token]
 S_TRAILING = lambda x, token:  ['trailing', None]
@@ -74,9 +75,9 @@ SCANNER = re.Scanner([
     (r"[a-zA-Z0-9._%+-]+@[a-zA-Z0-9.-]+\.[a-zA-Z]{2,4}", S_EMAIL_ADDR),
     (r"[0-9]+\.[0-9]+\.[0-9]+\.[0-9]", S_IP_ADDRESS),
     (r"-+[a-zA-Z0-9]+", S_OPTION),
-    (r"True|False|true|false|yes|no", S_BOOL), # matching single letters causes problems
     (r"[0-9]+", S_INT),
     (r"--", S_TRAILING),
+    (r"(True|False|true|false|yes|no|t|f|y|n)\b", S_BOOL),
     (r"[a-z\-]+", S_WORD),
     (r"\s", S_EMPTY),
     (r".+", S_STRING)
@@ -92,6 +93,8 @@ def match(tokens, of_type = None):
     # check the type (first element)
     if of_type:
         if not peek(tokens, of_type):
+            if of_type == 'option':
+                raise ArgumentError("Expecting an option, not a %s" % tokens[0][0])
             raise ArgumentError("Expecting '%s' type of argument not %s in tokens: %r.  Read the help." % 
                                (of_type, tokens[0][0], tokens))
 
@@ -277,10 +280,22 @@ def command_module(mod, command, options, ending="_command"):
     kwargs = determine_kwargs(function)
     ensure_defaults(options, kwargs)
     try:
-        return function(**options)
-    except TypeError, exc:
-        print "ERROR: ", exc
+        function(**options)
+    except TypeError as e:
+        tb = traceback.format_exc()
 
+        if not ("function(**options)" in tb.splitlines()[2]) and (len(tb.splitlines())/2 == len(tb.split("File"))):
+            print "//////////////////////////////////////////////////"
+            print tb
+            print "//////////////////////////////////////////////////"
+            raise Exception("traceback did not have expected contents, modargs is confused, please save the traceback printed above and report this bug")
+
+        if (len(tb.splitlines()) == 4):
+            print "Modargs Argument Error: %s" % e.message
+        else:
+            # This TypeError comes from within the application, modargs shouldn't catch it.
+            print tb
+            sys.exit(1)
 
 def available_help(mod, ending="_command"):
     """Returns the dochelp from all functions in this module that have _command
@@ -304,7 +319,7 @@ def help_for_command(mod, command, ending="_command"):
     """
 
     if command in available_commands(mod):
-        return mod.__dict__[command + ending].__doc__
+        return trim_docstring(mod.__dict__[command + ending].__doc__)
     else:
         return None
 
@@ -405,9 +420,9 @@ def help_text(prog, mod, default_command=None, on=False):
         text.append("Help for '%s'" % prog_on)
         text.append(dec)
         if command_help_text:
-            for line in command_help_text.split("\n"):
-                text.append(leading_spaces + line.strip())
-
+            for line in command_help_text.splitlines():
+                text.append(leading_spaces + line)
+            text.append("")
         fn = function_for(mod, on)
         kwargs = determine_kwargs(fn)
         kwdocs = determine_kwdocs(fn)
@@ -445,3 +460,78 @@ def help_text(prog, mod, default_command=None, on=False):
             text.append("%s%s\n\n" % (leading_spaces*2, kwdocs['kwargs']))
 
     return "\n".join(text)
+
+# From http://www.python.org/dev/peps/pep-0257/#handling-docstring-indentation
+def trim_docstring(docstring):
+    if not docstring:
+        return ''
+    # Convert tabs to spaces (following the normal Python rules)
+    # and split into a list of lines:
+    lines = docstring.expandtabs().splitlines()
+    # Determine minimum indentation (first line doesn't count):
+    indent = sys.maxint
+    for line in lines[1:]:
+        stripped = line.lstrip()
+        if stripped:
+            indent = min(indent, len(line) - len(stripped))
+    # Remove indentation (first line is special):
+    trimmed = [lines[0].strip()]
+    if indent < sys.maxint:
+        for line in lines[1:]:
+            trimmed.append(line[indent:].rstrip())
+    # Strip off trailing and leading blank lines:
+    while trimmed and not trimmed[-1]:
+        trimmed.pop()
+    while trimmed and not trimmed[0]:
+        trimmed.pop(0)
+    # Return a single string:
+    return '\n'.join(trimmed)
+
+def completion_command(prog, mod, default_command=None, on=False):
+    print completion_text(prog, mod, default_command, on)
+
+def completion_text(prog, mod, default_command=None, on=False):
+    cmds = available_commands(mod)
+
+    options_for_commands = []
+
+    for cmd in cmds:
+        fn = function_for(mod, cmd)
+        kwargs = determine_kwargs(fn)
+
+        keystring = " ".join(["-%s" % k for k in kwargs.keys()])
+
+        options_for_commands.append("       %s)" % cmd) # open case
+        options_for_commands.append("           local names=\"%s\"" % keystring)
+        options_for_commands.append("           COMPREPLY=( $(compgen -W \"${names}\" -- ${cur}) )")
+        options_for_commands.append("           return 0")
+        options_for_commands.append("             ;;") # close case
+
+    cmds_list = " ".join(cmds)
+
+    text_vars = {
+        'prog' : prog,
+        'cmds_list' : cmds_list,
+        'options_for_commands_text' : "\n".join(options_for_commands)
+    }
+
+    text = """_%(prog)s()
+{
+    local cur prev cmds base
+    COMPREPLY=()
+    cur="${COMP_WORDS[COMP_CWORD]}"
+    prev="${COMP_WORDS[COMP_CWORD-1]}"
+
+    case "${prev}" in
+%(options_for_commands_text)s
+        *)
+            ;;
+    esac
+
+    cmds="%(cmds_list)s"
+
+    COMPREPLY=($(compgen -W "${cmds}" -- ${cur}))
+    return 0
+}
+complete -F _%(prog)s %(prog)s""" % text_vars
+    return text
